"""
The module is about CRITT TPR-DB alignment.

The most important function in this module is `generate_alignment_pipelines`

It takes as the inputs ST and TT dataframes and returns alignment dictionary in the following format

{"P08_5_3":
            [   
                [[5, 9], [5, 6, 18]],  
                [[8], [11]]
            ]
}

Note that as for the Participant=08 , Text=5 , Segment=3 in the above 
example the source tokens [5,9] is aligned with target tokens [5,6,18]
and the source token [8] is aligned with target token [11] 

"""


import pandas as pd


def generate_orginal_alignment_list(target_dataframe):

    df = target_dataframe[['Id','TToken','Text', 'TTseg', 'Part', 'STid']]
    df["Part_Text_TTseg"] = df.apply(lambda x : str(x['Part']) + "_" + str(x['Text']) + "_" + str(x['TTseg']), axis=1)
    df = df[['Id', 'Part_Text_TTseg', 'STid','TToken']]
    df.columns = ['TTid', 'Part_Text_TTseg', 'STid','TToken']

    alignments_original={}

    for item in list(df["Part_Text_TTseg"].unique()):
        ptt = df[df["Part_Text_TTseg"] == item]
        alignments_original[item]=[]
        for value in list(ptt["STid"].unique()):
            if value != "---" :
                ttids = ptt[ptt["STid"]==value]
                st = [int(item) for item in value.split("+")]
                tt = list(ttids["TTid"])
                alignments_original[item].append([ st, tt ])

    return alignments_original


def get_offset_list(source_dataframe, target_dataframe):


    df_tt_filtered = target_dataframe[['Id','TToken','Text', 'TTseg', 'Part', 'STid']]
    df_tt_filtered["Part_Text_TTseg"] = df_tt_filtered.apply(lambda x : str(x['Part']) + "_" + str(x['Text']) + "_" + str(x['TTseg']), axis=1)
    df_tt_filtered = df_tt_filtered[['Id', 'Part_Text_TTseg', 'STid','TToken']]
    df_tt_filtered.columns = ['TTid', 'Part_Text_TTseg', 'STid','TToken']


    df_st_filtered = source_dataframe[['Id','SToken','Text', 'STseg', 'Part', 'TTid']]
    df_st_filtered["Part_Text_STseg"] = df_st_filtered.apply(lambda x : str(x['Part']) + "_" + str(x['Text']) + "_" + str(x['STseg']), axis=1)
    df_st_filtered = df_st_filtered[['Id', 'Part_Text_STseg', 'TTid','SToken']]
    df_st_filtered.columns = ['STid', 'Part_Text_STseg', 'TTid','SToken']

    offset_list={}

    for item in list(df_tt_filtered["Part_Text_TTseg"].unique()):
        pttt = df_tt_filtered[df_tt_filtered["Part_Text_TTseg"] == item]
        ptst = df_st_filtered[df_st_filtered["Part_Text_STseg"] == item]
        offset_list[item]=[]
        offset_list[item].append((ptst["STid"].min(),pttt["TTid"].min()))

    return offset_list


def generate_reindexed_alignment_list(alignments_original, offset_list):

    alignments_offset={}

    for item in alignments_original:
        alignments_offset[item]=[]
        st_offset = offset_list[item][0][0]
        tt_offset = offset_list[item][0][1]
        
        for align in alignments_original[item]:
            ss_items = []
            tt_items = []
            for left_item in align[0]:
                left_item -= st_offset
                ss_items.append(left_item)
            for right_item in align[1]:
                right_item -= tt_offset
                tt_items.append(right_item)

            alignments_offset[item].append([ss_items,tt_items])

    return alignments_offset


def generate_alignment_pipelines(df_st, df_tt):

    alignments_original = generate_orginal_alignment_list(df_tt)
    offset_list = get_offset_list(df_st, df_tt)
    alignments_offset =  generate_reindexed_alignment_list(alignments_original, offset_list)

    return alignments_offset


def get_alignment_offset(part, text, ttseg, alignments_offset):
    global_id = str(part) + "_" + str(text) + "_" + str(ttseg)

    for item in alignments_offset:
        if item == global_id:
            return alignments_offset[item]
    return None


# A module for calculating Incomplete Dependency Theory in a sentence (either in tokens or string format)

from lingx.core.lang_features import get_linguistic_features
from lingx.core.lang_features import aggregate_tokens
from lingx.core.lang_model import get_doc

def get_idt_complexity(input, nlp):

    idt_complexity=[]
    doc = get_doc(input, nlp)

    links , links_compact = get_linguistic_features(doc)
    

    for i in range(len(links_compact)):
        backward_links_above_index_i=[link for link in links if (link[1][0]<=i and link[0][0]>=(i+1))]
        forward_links_above_index_i=[link for link in links if (link[0][0]<=i and link[1][0]>=(i+1))]
        links_above_index_i=backward_links_above_index_i+forward_links_above_index_i
        idt_complexity.append([links[i][1][1], len(links_above_index_i)])

    return(idt_complexity)



def get_idt_score(input, nlp, result_format="token" , aggregation_type="sum"):
    """
    Argument `input`: 
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 

    Argument `result_format`: has values `token` or `segment`
    Argument `aggregation_type` : has values `sum` , `mean` or `max`

    """
    token = get_idt_complexity(input, nlp)

    if result_format == "token":
        return token, None
    elif result_format == "segment":
        return token, aggregate_tokens(token, aggregation_type)
    else:
        raise Exception("Set `result_format` to `token` or `segment`.")






import numpy as np
from lingx.core.lang_model import get_nlp_object
from lingx.metrics.bilingual.bcr import get_bcr_score

def test_bcr():

    nlp_en = get_nlp_object("en", use_critt_tokenization = True, package="partut")
    nlp_zh = get_nlp_object("zh", use_critt_tokenization = True)

    tokens_source = [['It', 'has', 'the', 'right', 'members', 'and', 'the', 'right', 'mandate', '.']]
    tokens_target = [['在', '今年', '的', '四月', '阿布特', '部长', '跟', '安倍', '部长', 
                    '同意', '为', '增强', '经济发展', '增强', '后', '合作', '，', '并且', 
                    '为', '太平洋', '地区', '的', '和平', '和', '稳定', '增强', '合作', '。']]


    source_target_alignments = [

        [[5, 9], [5, 6, 18]], 

        [[8], [11]]

        ]


    score = get_bcr_score(
                        nlp_source = nlp_en,
                        nlp_target = nlp_zh,
                        complexity_type = "idt_dlt", # "idt", "dlt", "idt_dlt"
                        input_source = tokens_source,
                        input_target = tokens_target,
                        source_target_alignments = source_target_alignments,
                        complexity_aggregation_function= "sum",    # max, mean, sum
                        first_aggregation_function= "sum",         # max, mean, sum
                        second_aggregation_function = "sum",       # max, mean, sum
                        robust=True,  # if set to `False` the error will be reported in case of alignment mismatch 
                        bcr_error_value=np.nan)  # if `robust` set to `True` the error will NOT be 
                                                    # reported and the `bcr_error_value` will be reported instead

    score_value = 8.7

    assert score_value >= score



# Manipulate TPRDB study-tables and perform different operations on them.

import pandas as pd
import numpy as np
import glob

from lingx.utils.lx import get_sentence_lx, get_sentence_mono_lingual
from lingx.utils.critt.aligner import get_alignment_offset

from lingx.metrics.bilingual.bcr import get_bcr_score


def readTPRDBtables(studies, table_type, verbose=0, path="/data/critt/tprdb/TPRDB/"):
    df = pd.DataFrame()
    
    for study in studies:
        if(verbose) : print("Reading: " + path + study + table_type)
        for fn in glob.glob(path + study + table_type):
            if(verbose > 1) : print("Reading: " + fn)
            df = pd.concat([df, pd.read_csv(fn, sep="\t", dtype=None)], ignore_index=True)
        
    return(df)


def convert_st2segment(df_st):
    analysis_st = (
                    pd.pivot_table(
                                    df_st,
                                    index=["Part","Text","STseg"], 
                                    values=["SToken"], 
                                    aggfunc=lambda x: '//'.join(x)
                                   )
                                    .reset_index()
                    )
    analysis_st["SToken"] = analysis_st["SToken"].apply(lambda x: [x.split("//")])
    
    return analysis_st



def convert_tt2segment(df_tt):
    analysis_tt = (
                    pd.pivot_table(
                                    df_tt,
                                    index=["Part","Text","TTseg"], 
                                    values=["TToken"], 
                                    aggfunc=lambda x: '//'.join(x)
                                   )
                                    .reset_index()
                    )
    analysis_tt["TToken"] = analysis_tt["TToken"].apply(lambda x: [x.split("//")])
    
    return analysis_tt




def expand_table_psycholingual(df_analysis, nlp, token_column="SToken"):

    df = df_analysis.copy()

    operation_list_st = [
                        ["idt","max","IDT_MAX_ST"],
                        ["idt","mean","IDT_MEAN_ST"],
                        ["idt","sum","IDT_SUM_ST"],
                        ["dlt","max","DLT_MAX_ST"],
                        ["dlt","mean","DLT_MEAN_ST"],
                        ["dlt","sum","DLT_SUM_ST"],
                        ["idt_dlt","max","IDT_DLT_MAX_ST"],
                        ["idt_dlt","mean","IDT_DLT_MEAN_ST"],
                        ["idt_dlt","sum","IDT_DLT_SUM_ST"],
    ]

    operation_list_tt = [
                        ["idt","max","IDT_MAX_TT"],
                        ["idt","mean","IDT_MEAN_TT"],
                        ["idt","sum","IDT_SUM_TT"],
                        ["dlt","max","DLT_MAX_TT"],
                        ["dlt","mean","DLT_MEAN_TT"],
                        ["dlt","sum","DLT_SUM_TT"],
                        ["idt_dlt","max","IDT_DLT_MAX_TT"],
                        ["idt_dlt","mean","IDT_DLT_MEAN_TT"],
                        ["idt_dlt","sum","IDT_DLT_SUM_TT"],
    ]


    if token_column == "SToken":
        operation_list = operation_list_st
    elif token_column == "TToken":
        operation_list = operation_list_tt
    else:
        raise Exception("Set `token_column` to one of the following values: `SToken` or `TToken`.")



    for item in operation_list:

        func = lambda x : get_sentence_lx(
                                            x,
                                            nlp,
                                            result_format="segment",
                                            complexity_type=item[0], 
                                            aggregation_type=item[1]
                                        )[1]

        df[item[2]]=df[token_column].apply(func)
        df[item[2]]=df[token_column].apply(func)

    return df


def expand_table_monolingual(df_analysis, nlp, token_column="SToken"):

    df = df_analysis.copy()

    operation_list_st = [
                        ["le","mean","LE_MEAN_ST"],
                        ["le","max","LE_MAX_ST"],
                        ["le","sum","LE_SUM_ST"],
                        ["mbn","mean","MBN_MEAN_ST"],
                        ["mbn","max","MBN_MAX_ST"],
                        ["mbn","sum","MBN_SUM_ST"]
    ]


    operation_list_tt = [
                        ["le","mean","LE_MEAN_TT"],
                        ["le","max","LE_MAX_TT"],
                        ["le","sum","LE_SUM_TT"],
                        ["mbn","mean","MBN_MEAN_TT"],
                        ["mbn","max","MBN_MAX_TT"],
                        ["mbn","sum","MBN_SUM_TT"]
    ]


    if token_column == "SToken":
        operation_list = operation_list_st
    elif token_column == "TToken":
        operation_list = operation_list_tt
    else:
        raise Exception("Set `token_column` to one of the following values: `SToken` or `TToken`.")


    for item in operation_list:

        func = lambda x : get_sentence_mono_lingual(x, nlp, mono_lingual_type=item[0], aggregation_type=item[1])
        
        df[item[2]]=df[token_column].apply(func)

    return df


def expand_table_bilingual(analysis_st_tt, nlp_source, nlp_target, robust=True, bcr_error_value=0):

    df = analysis_st_tt.copy()

    bcr_aggregation_list = ["sum", "max", "mean"]
    lx_aggregation_list = ["sum"]

    for first_agg in bcr_aggregation_list:
        for second_agg in bcr_aggregation_list:
            for lx in lx_aggregation_list:

                func = lambda x : get_bcr_score(
                                            nlp_source,
                                            nlp_target,
                                            complexity_type = "idt_dlt", # "idt", "dlt", "idt_dlt"
                                            input_source = x["SToken"],
                                            input_target = x["TToken"],
                                            source_target_alignments = x["Alignment"],
                                            complexity_aggregation_function= lx,         # max, mean, sum
                                            first_aggregation_function= first_agg,       # max, mean, sum
                                            second_aggregation_function = second_agg,    # max, mean, sum
                                            robust=robust,
                                            bcr_error_value=bcr_error_value
                                    )

                label = "BCR" +  "_" + lx.upper() +"_" + first_agg.upper() + "_" + second_agg.upper()
                df[label] = df.apply(func, axis=1)

    return df




def merge_st_tt(analysis_st, analysis_tt, alignments_offset):

    analysis_st_renamed = analysis_st.rename(columns = {"STseg":"TTseg"})

    analysis_tt_extended = pd.merge(analysis_tt,
                                    analysis_st_renamed,
                                    on=["Part", "Text", "TTseg"], 
                                    how='left')

    analysis_tt_extended["Alignment"] = analysis_tt_extended.apply(
                                                                    lambda x: get_alignment_offset(
                                                                                x["Part"], x["Text"], x["TTseg"], 
                                                                                alignments_offset),
                                                                    axis=1
                                                                    )

    return analysis_tt_extended



def expand_table_error(analysis_st_tt, error_file_path):

    errors = pd.read_csv(error_file_path, index_col=False)

    analysis_st_tt["SessionSeg"] = analysis_st_tt.apply(lambda x :x["Part"] + "_Ist" + str(x["Text"])+ "_" + str(x["TTseg"]), axis=1)
    analysis_st_tt = pd.merge(analysis_st_tt,
                                    errors,
                                    on=["SessionSeg"],
                                    how='left')
    return analysis_st_tt


# A module for calculating Left Embeddedness in a sentence

import numpy as np

from lingx.core.lang_features import is_aux_a_main_verb, is_a_main_verb
from lingx.core.lang_model import get_doc


def get_le_from_verb(verb_indicator, aggregator="sum"):

    drop_last_score = True
    false_place_holders = "".join(str(e) for e in verb_indicator).replace("False","X").split("True")

    if false_place_holders[-1] == '':
        false_place_holders.pop()
        drop_last_score = False
    if false_place_holders[0] == '':
        false_place_holders.pop(0)

    non_verb_counts = [len(item) for item in false_place_holders]

    if drop_last_score :
        non_verb_counts = non_verb_counts[:-1]

    if len(non_verb_counts) > 0:
        if aggregator == "sum":
            left_embeddedness_score=np.sum(non_verb_counts)
        elif aggregator == "mean":
            left_embeddedness_score=np.mean(non_verb_counts)
        elif aggregator == "max":
            left_embeddedness_score=np.max(non_verb_counts)
    else:
        left_embeddedness_score = 0 # The default value in case we have no VERB

    return left_embeddedness_score, non_verb_counts

def get_le_score(input, nlp, aggregator="sum"):
    """
    Argument `input`: 
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 

    Argument `result_format`: has values `token` or `segment`
    Argument `aggregation_type` : has values `sum` , `mean` or `max`

    """

    doc = get_doc(input, nlp, ignore_empty_input=True)

    if doc==[]:
        return 0

    doc_verb_indicators=[]
    for sent in doc.sentences:
        verb_indicator=[]
        for word in sent.words:
                if is_aux_a_main_verb(word) or is_a_main_verb(word):
                    verb_indicator.append(True)
                else:
                    verb_indicator.append(False)     
        doc_verb_indicators.append(verb_indicator)

    score_lists=[]
    saved_score_lists=[]
    for verb_indicator in doc_verb_indicators:
        le_score = get_le_from_verb(verb_indicator,aggregator)
        score_lists.append(le_score[0])
        saved_score_lists.append(le_score)

    if aggregator == "sum":
        left_embeddedness_score=np.sum(score_lists)
    elif aggregator == "mean":
        left_embeddedness_score=np.mean(score_lists)
    elif aggregator == "max":
        left_embeddedness_score=np.max(score_lists)

    # possible values to return left_embeddedness_score, saved_score_lists, doc_verb_indicators
    return left_embeddedness_score


# A module for calculating Bilingual Complexity Ratio in a give source and target texts

import numpy as np

from lingx.utils.lx import get_sentence_lx
from lingx.core.lang_model import get_doc



def has_misalignment(input_source, input_target, source_target_alignments):

    # set `side` as 0 for verification of source indexes
    # set `side` as 1 for verification of target indexes
    def report_index_error(tokens_input, source_target_alignments, side):   
        error_tag = False
        len_list=[]

        if isinstance(tokens_input, str):
            tokens_input = [tokens_input.split(" ")]


        for item in tokens_input:
            len_list.append(len(item))
        length=sum(len_list)


        for item in source_target_alignments:
            for number in item[side]:
                if number >= length:
                    error_tag = True
                    break

        return error_tag

    source_err_flag = report_index_error(input_source, source_target_alignments, side=0)
    target_err_flag = report_index_error(input_target, source_target_alignments, side=1)

    return (source_err_flag or target_err_flag)





def get_bcr_score_(
                    tokens_source_complexity,
                    tokens_target_complexity,
                    source_target_alignments,
                    first_aggregation_function= "sum",    # max, mean, sum
                    second_aggregation_function = "sum"   # max, mean, sum
                    ):

    scores=[]
    for item in source_target_alignments:
        left_scores=[]
        right_scores=[]

        for left_item in item[0]:
            left_scores.append(tokens_source_complexity[left_item][1])

        for right_item in item[1]:
            right_scores.append(tokens_target_complexity[right_item][1])

        if first_aggregation_function == "sum":
            st_score=np.sum(left_scores)
            tt_score=np.sum(right_scores)
        elif first_aggregation_function == "mean":
            st_score=np.mean(left_scores)
            tt_score=np.mean(right_scores)
        elif first_aggregation_function == "max":
            st_score=np.max(left_scores)
            tt_score=np.max(right_scores)
        else:
            st_score=np.sum(left_scores)
            tt_score=np.sum(right_scores)

        scores.append(tt_score/st_score)


    if second_aggregation_function == "sum":
        bcr_score=np.sum(scores)
    elif second_aggregation_function == "mean":
        bcr_score=np.mean(scores)
    elif second_aggregation_function == "max":
        if scores!= []:
            bcr_score=np.max(scores)
        else:
            bcr_score=0
    else:
        bcr_score=np.sum(scores)

    return bcr_score





def get_bcr_score(
                    nlp_source,
                    nlp_target,
                    complexity_type, # IDT, DLT, IDT+DLT
                    input_source,
                    input_target,
                    source_target_alignments,
                    complexity_aggregation_function= "sum",    # max, mean, sum
                    first_aggregation_function= "sum",         # max, mean, sum
                    second_aggregation_function = "sum",       # max, mean, sum
                    robust=True,
                    bcr_error_value=0):




    has_error = has_misalignment(input_source, input_target, source_target_alignments)

    if has_error and robust:
        return (bcr_error_value)
    elif has_error and not robust:
        raise Exception("A misalignmet exist in input_source={}, input_target={} and alignment={}.".format(input_source, input_target, source_target_alignments))


    tokens_source_complexity , _ =  get_sentence_lx(
                                                input_source,
                                                nlp_source,
                                                result_format="token",
                                                complexity_type=complexity_type, 
                                                aggregation_type=complexity_aggregation_function
                                                )

    tokens_target_complexity , _ =  get_sentence_lx(
                                                input_target,
                                                nlp_target,
                                                result_format="token",
                                                complexity_type=complexity_type, 
                                                aggregation_type=complexity_aggregation_function
                                                )


    bcr_score = get_bcr_score_(
                                tokens_source_complexity,
                                tokens_target_complexity,
                                source_target_alignments,
                                first_aggregation_function,
                                second_aggregation_function
                               )
    
    return bcr_score





from lingx.core.lang_model import get_nlp_object
from lingx.metrics.monolingual.le import get_le_score

def test_le():

    nlp_en = get_nlp_object("en", use_critt_tokenization = True, package="partut")

    input = [["The", "reporter", "who", "the", "senator", "who", "John", "met", "attacked"], ["disliked", "the", "editor", "."]]

    score = get_le_score(
                        input, 
                        nlp=nlp_en, 
                        aggregator="sum")  # choose `sum`, `max` or `mean`

    score_value = 7

    assert score == score_value



from lingx.core.lang_model import get_nlp_object

from lingx.metrics.psycholingual.idt import get_idt_score
from lingx.metrics.psycholingual.dlt import get_dlt_score
from lingx.metrics.psycholingual.idt_dlt import get_idt_dlt_score
from lingx.metrics.monolingual.le import get_le_score
from lingx.metrics.monolingual.mbn import get_mbn_score

from lingx.core.lang_features import aggregate_tokens


def compare_sentences_lx(
                        sent_a, 
                        sent_b,
                        nlp,
                        complexity_type="idt_dlt", 
                        aggregation_type="sum"):


    if complexity_type == "idt":
        result_a = get_idt_score(sent_a, nlp, result_format="segment" , aggregation_type=aggregation_type)
        result_b = get_idt_score(sent_b, nlp, result_format="segment" , aggregation_type=aggregation_type)

    elif complexity_type == "dlt":
        result_a = get_dlt_score(sent_a, nlp, result_format="segment" , aggregation_type=aggregation_type)
        result_b = get_dlt_score(sent_b, nlp, result_format="segment" , aggregation_type=aggregation_type)

    if complexity_type == "idt_dlt":
        result_a = get_idt_dlt_score(sent_a, nlp, result_format="segment" , aggregation_type=aggregation_type)
        result_b = get_idt_dlt_score(sent_b, nlp, result_format="segment" , aggregation_type=aggregation_type)

    return (result_b[1] > result_a[1], result_a[0], result_b[0])




def get_sentence_lx(
                    input,
                    nlp,
                    result_format="token",
                    complexity_type="idt_dlt", 
                    aggregation_type="sum"):

    if complexity_type == "idt":
        score = get_idt_score(input, nlp, result_format=result_format , aggregation_type=aggregation_type)
    elif complexity_type == "dlt":
        score = get_dlt_score(input, nlp, result_format=result_format , aggregation_type=aggregation_type)
    elif complexity_type == "idt_dlt":
        score = get_idt_dlt_score(input, nlp, result_format=result_format , aggregation_type=aggregation_type)
    else:
        raise Exception("Set `complexity_type` to one of the following values: `idt` , `dlt` or `idt_dlt`.")
    
    return score



def get_sentence_mono_lingual(
                                input, nlp,
                                mono_lingual_type="le",  # 'le' or 'mbn'
                                aggregation_type="sum"):

    if mono_lingual_type == "le":
        score = get_le_score(input, nlp, aggregation_type)
    elif mono_lingual_type == "mbn":
        score = get_mbn_score(input, nlp, aggregation_type)
    else:
        raise Exception("Set `mono_lingual_type` to one of the following values: `le` or `mbn`.")
    
    return score



import numpy as np
from treelib import Tree, Node

def get_linguistic_features(doc):
    links=[]
    links_compact=[] 
    sent_counter=0
    acc=0

    for sent in doc.sentences:
        sent_counter+=acc
        acc=len(sent.words)
        for word in sent.words:
            if word.head > 0:
                links.append([[sent_counter+word.head-1, sent.words[word.head-1].text, sent.words[word.head-1].upos], [sent_counter+word.id-1, word.text, word.upos]])
                links_compact.append([sent_counter+word.head-1, sent_counter+word.id-1])
                    
            else:
                links.append([[sent_counter+word.id-1, word.text, word.upos], [sent_counter+word.id-1, word.text, word.upos]])
                links_compact.append([sent_counter+word.id-1, sent_counter+word.id-1])

    return links, links_compact


def aggregate_tokens(tokens_complexity, aggregation_type="sum"):

    list_tokens_complexity=[token[1] for token in tokens_complexity]

    if len(list_tokens_complexity)>0:

        if aggregation_type=="sum":
            return np.sum(list_tokens_complexity)
        elif aggregation_type=="min":
            return np.min(list_tokens_complexity)
        elif aggregation_type=="max":
            return np.max(list_tokens_complexity)
        elif aggregation_type=="mean":
            return np.average(list_tokens_complexity)
        else:
            raise Exception("Set `aggregation_type` to one of the values `sum`, `mean`, `min` or `max`.")

    else:
        return 0

def is_aux_a_main_verb(word):
    if word.upos=="AUX" and  word.deprel=="aux":
        return False
    elif word.upos=="AUX" and  word.deprel!="aux":
        return True
    else:
        return False

def is_a_main_verb(word):
    if word.upos=="VERB":
        return True
    else:
        return False

def is_a_noun(word):
    if word.upos=="NOUN" or word.upos=="PROPN":
        return True
    else:
        return False


def convert_dep2tree(list_pairs):
    tree = Tree()
    tree.create_node("root", 0)
    pairs = list_pairs.copy()

    while (len(pairs)):
        for item in tree.leaves():
            planned_to_be_deleted=[]
            for pair in pairs:
                if item.identifier == pair[0]:
                    tree.create_node(pair[3], pair[1], parent=pair[0])
                    planned_to_be_deleted.append(pair)
            
            for pair in planned_to_be_deleted:
                pairs.remove(pair)
    return(tree)


# A module for calculating Distance Locality Theory in a sentence (either in tokens or string format)

from lingx.core.lang_features import get_linguistic_features
from lingx.core.lang_features import aggregate_tokens
from lingx.core.lang_model import get_doc


acceptable_discourse_referents = ["PROPN", "NOUN", "VERB"]


def get_longest_link(links, links_compact):
    backward_longest_links=[]

    for link in links:
        for indx in range(0,link[1][0]):
            if ([link[1][0], indx] in links_compact):
                backward_longest_links.append([link[1][0], indx])
                break

    return(backward_longest_links)


def get_dlt_complexity(input, nlp):
    """
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 
    """
    doc = get_doc(input, nlp)

    links , links_compact = get_linguistic_features(doc)
    backward_longest_links=get_longest_link(links, links_compact)

    dlt_complexity_partial=[]
    for link in backward_longest_links:
        accepted_pos=[]
        for indx in range(link[0]-1,link[1]-1,-1):
            pos=links[indx][1][2]
            if pos in acceptable_discourse_referents: 
                accepted_pos.append(pos)
        dlt_complexity_partial.append([link[0], len(accepted_pos)])
    
    dlt_complexity=[]
    token_indices=[indx for indx, _ in dlt_complexity_partial]

    for i in range(len(links_compact)):
        if i not in token_indices:
            dlt_complexity.append([links[i][1][1], 0])
        else:
            dlt_complexity.append([links[i][1][1], *[pair[1] for pair in dlt_complexity_partial if pair[0]==i]])
    
    return(dlt_complexity)




def get_dlt_score(input, nlp, result_format="token" , aggregation_type="sum"):
    """
    Argument `input`: 
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 

    Argument `result_format`: has values `token` or `segment`
    Argument `aggregation_type` : has values `sum` , `mean` or `max`

    """
    token = get_dlt_complexity(input, nlp)

    if result_format == "token":
        return token, None
    elif result_format == "segment":
        return token, aggregate_tokens(token, aggregation_type)
    else:
        raise Exception("Set `result_format` to `token` or `segment`.")


from lingx.utils import download_lang_models
from lingx.core.lang_model import get_nlp_object
from lingx.utils.lx import compare_sentences_lx

nlp_en = get_nlp_object("en", use_critt_tokenization = False, package="partut")
nlp_zh = get_nlp_object("zh", use_critt_tokenization = False)


sentences = [
    ["The reporter who the senator who I met attacked disliked the editor.",
    "The reporter who the senator who John met attacked disliked the editor."],
    ["The reporter who sent the photographer to the editor hoped for a good story.",
    "The reporter who the photographer sent to the editor hoped for a good story."],
    ["The horse raced past the barn.", 
    "The horse raced past the barn fell."],
    ["The dog that chased the cat that saw the rat barked.", 
    "The cheese that the rat that the cat saw ate stank."],
    ["Jack was surprised that two plus two equals four.", 
    "The fact that two plus two equals four surprised Jack."],
    ["Ingrid was astonished that Jack was surprised that two plus two equals four.", 
    "The fact that that two plus two equals four surprised Jack astonished Ingrid."],
    ["Frank was bothered that Ingrid was astonished that Jack was surprised that two plus two equals four.", 
    "The fact that that that that that two plus two equals four surprised Jack astonished Ingrid bothered Frank."],
    ["John gave Bill the painting that Mary hated.", 
    "John gave the painting that Mary hated to Bill."],
    ["That two plus two equals four surprised Jack.",
    "That that two plus two equals four surprised Jack astonished Ingrid."]
]

for sent in sentences:
    result = compare_sentences_lx(
                        sent[0], 
                        sent[1],
                        nlp=nlp_en,
                        complexity_type="idt_dlt", 
                        aggregation_type="sum")

    print()
    print(f"Sentence 1 : {result[1]}")
    print(f"Sentence 2 : {result[2]}")
    print(f"Fx(Sentence 2) > Fx(Sentence 1) : {result[0]}")
    print()


# A module for calculating number of Modifiers Before Nouns in a sentence

import numpy as np
from treelib import Tree, Node


from lingx.core.lang_features import is_a_noun, convert_dep2tree
from lingx.core.lang_model import get_doc



def get_mbn_from_noun_indicator(noun_indicator, tree, aggregator="mean", ploraity = False):
    scores = []

    if len(noun_indicator) > 1:
        for index in range(len(noun_indicator)-1):
            fst = noun_indicator[index]
            snd = noun_indicator[index+1]

            if tree.is_ancestor(fst, snd):
                score = fst-snd
            elif tree.is_ancestor(snd, fst):
                score = snd-fst
            else:
                score = 0

            if ploraity:
                scores.append(score)
            else:
                scores.append(abs(score))
    else:
        scores.append(0)

    if len(noun_indicator) > 0:
        if aggregator == "sum":
            nested_np_modifiers_score=np.sum(scores)
        elif aggregator == "mean":
            nested_np_modifiers_score=np.mean(scores)
        elif aggregator == "max":
            nested_np_modifiers_score=np.max(scores)
    else:
        nested_np_modifiers_score = 0 # The default value in case we have no NOUN

    return nested_np_modifiers_score, scores


def get_mbn_score(input, nlp, aggregator="sum", ploraity = False):
    """
    Argument `input`: 
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 

    Argument `result_format`: has values `token` or `segment`
    Argument `aggregation_type` : has values `sum` , `mean` or `max`

    """

    doc = get_doc(input, nlp, ignore_empty_input=True)

    if doc==[]:
        return 0

    sent_list_pairs = []
    for sent in doc.sentences:
        list_pairs=[]
        noun_indicator = []

        for word in sent.words:
            
            if is_a_noun(word):
                noun_indicator.append(word.id)

            if word.head == 0:
                list_pairs.append([word.head, word.id, "root", word.text])
            else:
                list_pairs.append([word.head, word.id, sent.words[word.head-1].text, word.text])
        sent_list_pairs.append([list_pairs,noun_indicator])


    score_lists=[]
    saved_score_lists=[]
    for list_pairs in sent_list_pairs:
        dep_tree = convert_dep2tree(list_pairs[0])
        le_score = get_mbn_from_noun_indicator(list_pairs[1], dep_tree, aggregator, ploraity)
        score_lists.append(le_score[0])
        saved_score_lists.append(le_score)


    if aggregator == "sum":
        nested_np_modifiers_score=np.sum(score_lists)
    elif aggregator == "mean":
        nested_np_modifiers_score=np.mean(score_lists)
    elif aggregator == "max":
        nested_np_modifiers_score=np.max(score_lists)

    # possible return values for further tests are nested_np_modifiers_score, saved_score_lists, sent_list_pairs
    return nested_np_modifiers_score


    


import stanza

def download_stanza_model(stanza_model, package="default"):
    stanza.download(stanza_model, package=package)


def get_nlp_object(stanza_model, use_critt_tokenization = True, package="default"):

    if use_critt_tokenization:
        nlp = stanza.Pipeline(stanza_model, package=package, processors='tokenize, pos, lemma, depparse', tokenize_pretokenized=True)
    else:
        nlp = stanza.Pipeline(stanza_model, package=package,  processors='tokenize, pos, lemma, depparse', tokenize_pretokenized=False)

    return nlp, use_critt_tokenization
    

# In order to avoid following stanza libraray error this subroutine should be run
# AssertionError: id and text should be included for the token
def convert_blank2space(input):

    tokens_list=[]
    for items in input:
        tokens=[]
        for token in items:
            if token == "":
                tokens.append(" ")
            else:
                tokens.append(token)
        tokens_list.append(tokens)
    
    return tokens_list


# `use_critt_tokenization` argument is problematic in general 
# the following procedure handles the complexity
def get_doc(input, nlp, ignore_empty_input=False):
    
    ignore_empty_list = [
                            "", 
                            [], 
                            [[]]
                        ]

    if ignore_empty_input and input in ignore_empty_list:
            return []

    if nlp[1] and isinstance(input, list):
        input = convert_blank2space(input)
        doc = nlp[0](input)
    elif nlp[1] and (not isinstance(input, list)):
        raise Exception("You have set `use_critt_tokenization` to True but not used real tokens (in list_of_list format) as segment!")
    elif (not nlp[1]) and isinstance(input, str):
        doc = nlp[0](input)
    elif (not nlp[1]) and (not isinstance(input, str)):
        raise Exception("You have set `use_critt_tokenization` to False but not used real string as sentence!")

    return(doc)







from setuptools import setup, find_packages

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name="lingx",
    version="0.1.5",
    description="A library for introducing state-of-the-art metrics on measuring linguistic complexity",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Mehdi Mirzapour",
    author_email="mehdi.mirzapour@gmail.com",
    url="https://github/ContentSide/lingx",
    packages=find_packages(exclude="tests"),  # same as name
    license="MIT",
    install_requires=required,
    include_package_data=True,
    python_requires=">=3.6"
)


from lingx.core.lang_model import get_nlp_object
from lingx.utils.lx import get_sentence_lx

def test_dlt():

    nlp_en = get_nlp_object("en", use_critt_tokenization = True, package="partut")

    input = [["The", "reporter", "who", "the", "senator", "who", "John", "met", "attacked"], ["disliked", "the", "editor", "."]]

    tokens_scores_list, aggregated_score = get_sentence_lx(
                                                        input,
                                                        nlp_en,
                                                        result_format="segment",
                                                        complexity_type="dlt", 
                                                        aggregation_type="sum")  # choose `sum`, `max` or `mean`

    tokens_scores_list_value = [['The', 0], ['reporter', 0], ['who', 0], ['the', 0], 
                                ['senator', 0], ['who', 0], ['John', 0], ['met', 2], 
                                ['attacked', 0], ['disliked', 0], ['the', 0], ['editor', 0], ['.', 0]]
    aggregated_score_value = 2

    assert tokens_scores_list == tokens_scores_list_value
    assert aggregated_score == aggregated_score_value



# A module for calculating Incomplete Dependency and Distance Locality Theories in a sentence (either in tokens or string format)

from lingx.metrics.psycholingual.idt import get_idt_complexity
from lingx.metrics.psycholingual.dlt import get_dlt_complexity
from lingx.core.lang_features import aggregate_tokens



def calc_idt_dlt_complexity(self, tokens):
    tokens=[" " if t=="" else t for t in tokens]
    combined_complexity=[]
    
    idt_calc=self.calc_idt_complexity(tokens)
    dlt_calc=self.calc_dlt_complexity(tokens)

    for pair in zip(idt_calc,dlt_calc):
        combined_complexity.append([pair[0][0], pair[0][1]+pair[1][1]]) 

    return(combined_complexity)


def get_idt_dlt_complexity(input, nlp):

    combined_complexity = []
    
    idt = get_idt_complexity(input, nlp)
    dlt = get_dlt_complexity(input, nlp)

    for pair in zip(idt, dlt):
        combined_complexity.append([pair[0][0], pair[0][1]+pair[1][1]]) 

    return(combined_complexity)


def get_idt_dlt_score(input, nlp, result_format="token" , aggregation_type="sum"):
    """
    Argument `input`: 
    If input type is tokens the input should be in this format 
        [['token_1_of_sent_1', 'token_2_of_sent_1', ...],['token_1_of_sent_2', 'token_2_of_sent_2', ...]]
        Example : 
        [['This', 'is', 'token.ization', 'done', 'my', 'way!'], ['Sentence', 'split,', 'too!']])  

    If input type is string the input should be a simple standard python string 

    Argument `result_format`: has values `token` or `segment`
    Argument `aggregation_type` : has values `sum` , `mean` or `max`

    """
    token = get_idt_dlt_complexity(input, nlp)

    if result_format == "token":
        return token, None
    elif result_format == "segment":
        return token, aggregate_tokens(token, aggregation_type)
    else:
        raise Exception("Set `result_format` to `token` or `segment`.")